import logging

from flask import current_app, request

from slackminion.exceptions import DuplicateCommandError
from slackminion.slack import SlackChannel


class BaseCommand(object):
    def __init__(self, method):
        self.method = method
        self.help = method.__doc__

    def execute(self, *args, **kwargs):
        return self.method(*args, **kwargs)


class PluginCommand(BaseCommand):
    def __init__(self, method):
        super(PluginCommand, self).__init__(method)
        self.acl = method.acl
        self.admin_only = method.admin_only
        self.is_subcmd = method.is_subcmd
        self.while_ignored = method.while_ignored


class WebhookCommand(BaseCommand):
    def __init__(self, method, form_params):
        super(WebhookCommand, self).__init__(method)
        self.form_params = form_params

    def execute(self):
        args = {}
        form_params = self.form_params
        if isinstance(self.form_params, basestring):
            form_params = [self.form_params]
        if form_params is not None:
            for p in form_params:
                if p in request.form:
                    args[p] = request.form[p]
        return self.method(**args)


class MessageDispatcher(object):
    def __init__(self):
        self.log = logging.getLogger(type(self).__name__)
        self.commands = {}
        self.ignored_channels = []
        self.ignored_events = ['message_replied', 'message_changed']

    def push(self, message):
        """
        Takes a SlackEvent, parses it for a command, and runs against registered plugin
        """
        if self._ignore_event(message):
            return None, None
        args = self._parse_message(message)
        self.log.debug("Searching for command using chunks: %s", args)
        cmd, msg_args = self._find_longest_prefix_command(args)
        if cmd is not None:
            if message.user is None:
                self.log.debug("Discarded message with no originating user: %s", message)
                return None, None
            sender = message.user.username
            if message.channel is not None:
                sender = "#%s/%s" % (message.channel.name, sender)
            self.log.info("Received from %s: %s, args %s", sender, cmd, msg_args)
            f = self._get_command(cmd, message.user)
            if f:
                if self._is_channel_ignored(f, message.channel):
                    self.log.info("Channel %s is ignored, discarding command %s", message.channel, cmd)
                    return '_ignored_', ""
                return cmd, f.execute(message, msg_args)
            return '_unauthorized_', "Sorry, you are not authorized to run %s" % cmd
        return None, None

    def _ignore_event(self, message):
        """
        message_replied event is not truly a message event and does not have a message.text
        don't process such events

        commands may not be idempotent, so ignore message_changed events.
        """
        if hasattr(message, 'subtype') and message.subtype in self.ignored_events:
            return True
        return False

    def _parse_message(self, message):
        args = message.text.split(' ')
        return args

    def register_plugin(self, plugin):
        """Registers a plugin and commands with the dispatcher for push()"""
        self.log.info("Registering plugin %s", type(plugin).__name__)
        self._register_commands(plugin)
        plugin.on_load()

    def _register_commands(self, plugin):
        for name in dir(plugin):
            method = getattr(plugin, name)
            if callable(method) and hasattr(method, 'is_cmd'):
                commands = [method.cmd_name]
                if method.aliases is not None:
                    aliases = method.aliases
                    if isinstance(method.aliases, basestring):
                        aliases = [method.aliases]
                    for alias in aliases:
                        commands.append(alias)

                for cmd_name in commands:
                    cmd = '!' + cmd_name
                    if cmd in self.commands:
                        raise DuplicateCommandError(cmd_name)
                    self.log.info("Registered command %s", type(plugin).__name__ + '.' + cmd_name)
                    self.commands[cmd] = PluginCommand(method)
            elif callable(method) and hasattr(method, 'is_webhook'):
                self.log.info("Registered webhook %s", type(plugin).__name__ + '.' + name)
                webhook = WebhookCommand(method, method.form_params)
                with plugin._bot.webserver.app.app_context():
                    current_app.add_url_rule(method.route, method.__name__, webhook.execute, methods=[method.method])

    def ignore(self, channel):
        if isinstance(channel, SlackChannel):
            channel = channel.name
        if channel not in self.ignored_channels:
            self.ignored_channels.append(channel)
            return True
        return False

    def unignore(self, channel):
        if isinstance(channel, SlackChannel):
            channel = channel.name
        if channel in self.ignored_channels:
            self.ignored_channels.remove(channel)
            return True
        return False

    def _find_longest_prefix_command(self, args):
        num_parts = len(args)
        while num_parts > 0:
            cmd = ' '.join(args[0:num_parts])
            if cmd in self.commands:
                return cmd, args[num_parts:]
            num_parts -= 1
        return None, None

    def _get_command(self, cmd, user):
        can_run_cmd = True
        if hasattr(self, 'auth_manager'):
            can_run_cmd = self.auth_manager.admin_check(self.commands[cmd], user)
            if can_run_cmd:
                can_run_cmd = self.auth_manager.acl_check(self.commands[cmd], user)
        if not can_run_cmd:
            self.log.info("User %s is not authorized to run %s", user.username, cmd)
            return None
        return self.commands[cmd]

    def _is_channel_ignored(self, cmd, channel):
        channel_ignored = False
        if channel.name in self.ignored_channels:
            channel_ignored = not cmd.while_ignored
        return channel_ignored



from pkgutil import extend_path
from slackminion.plugin.base import BasePlugin

__path__ = extend_path(__path__, __name__)


class BaseStateHandler(BasePlugin):
    def on_load(self):
        # Don't save this plugin's state during save_state()
        self._dont_save = True
        self._state_handler = True

    def load_state(self):
        pass

    def save_state(self, state):
        pass



import os

from slackminion.plugins.core import version
try:
    from slackminion.plugins.core import commit
except ImportError:
    commit = 'HEAD'

from . import BaseStateHandler


class FileStateHandler(BaseStateHandler):
    def load_state(self):
        with open(os.path.join(self.config['data_dir'], 'state.json'), 'rb') as f:
            state = f.read()
        return state

    def save_state(self, state):
        with open(os.path.join(self.config['data_dir'], 'state.json'), 'wb') as f:
            f.write(state)



class SlackRoomAttribute(object):
    """Represents a room attribute, such as name, creator, topic, etc"""
    def __init__(self):
        self.data = None

    def __get__(self, instance, owner):
        if self.data is None:
            instance._load_extra_attributes()
        return self.data

    def __set__(self, instance, value):
        self.data = value


class SlackRoomTopic(SlackRoomAttribute):
    """Extension of SlackRoomAttribute, makes API call to set topic when __set__ is called"""
    def __set__(self, instance, value):
        prev_value = self.data
        super(SlackRoomTopic, self).__set__(instance, value)
        if prev_value is not None and prev_value != value:
            instance.set_topic(value)



import logging


class SlackRoomIMBase(object):
    def __init__(self, id, sc=None, **kwargs):
        """Base class for rooms (channels, groups) and IMs"""
        self.id = id
        self._sc = sc
        self.logger = logging.getLogger(type(self).__name__)
        self.logger.setLevel(logging.DEBUG)



import logging


class SlackUser(object):
    """Represents a Slack user"""
    def __init__(self, id, sc=None):
        self.id = id
        self._sc = sc
        self._username = None
        self.logger = logging.getLogger(type(self).__name__)
        self.logger.setLevel(logging.DEBUG)

    @property
    def username(self):
        if self._username is None and self._sc is not None:
            resp = self._sc.server.users.find(self.id)
            if resp is not None:
                self._username = resp.name
        return self._username

    @property
    def userid(self):
        self.logger.warn('Use of userid is deprecated, use id instead')
        return self.id

    @staticmethod
    def get_user(sc, username):
        resp = sc.server.users.find(username)
        if resp is None:
            return None
        user = SlackUser(resp.id, sc)
        return user

    def __str__(self):
        return '<@%s|%s>' % (self.id, self.username)

    def __repr__(self):
        return self.id



from functools import wraps

from slackminion.plugin import cmd
from slackminion.plugin.base import BasePlugin

from . import version
try:
    from . import commit
except ImportError:
    commit = 'HEAD'


def user_mgt_command(f):
    @wraps(f)
    def wrapper(self, msg, args):
        if len(args) < 2:
            return "Usage: !%s *acl_name* *username*" % f.__name__.replace('_', ' ')
        name, user = args
        return f(self, name, user)
    return wrapper


def acl_mgt_command(f):
    @wraps(f)
    def wrapper(self, msg, args):
        if len(args) < 1:
            return "Usage: !%s *acl_name*" % f.__name__.replace('_', ' ')
        name = args[0]
        return f(self, name)
    return wrapper


class AuthManager(BasePlugin):
    """Basic Authorization Plugin"""
    def on_load(self):

        # Hook into the dispatcher to receive acl checks
        setattr(self._bot.dispatcher, 'auth_manager', self)

        # Setup default ACL
        # ACL Rule Ordering (first match)
        # Allow Rule --> Deny Rule --> Allow All
        self._acl = {
            '*': {
                'allow': [],
                'deny': []
            },
        }

        return super(AuthManager, self).on_load()

    @cmd(admin_only=True)
    def acl(self, msg, args):
        """ACL Management.

        Usage:
        !acl _action_ [args]

        Actions:
        new _acl_ - Create a new ACL
        delete _acl_ - Delete an ACL

        allow _acl_ _user_ - Add user to the acl allow block
        deny _acl_ _user_ - Add user to the acl deny block
        remove _acl_ _user_ - Remove user from acl allow and deny blocks

        show - Show all defined ACLs
        show _acl_ - Show allow and deny blocks of specified ACL
        """
        if len(args) == 0:
            return "Usage: !acl show OR !acl _action_ _args_"

        valid_actions = ['allow', 'deny', 'remove', 'show', 'new', 'delete']
        return "Valid actions: %s" % ', '.join(valid_actions)

    @cmd(admin_only=True)
    @acl_mgt_command
    def acl_new(self, name):
        if self.create_acl(name):
            return "Created new acl '%s'" % name
        return "ACL '%s' already exists" % name

    @cmd(admin_only=True)
    @acl_mgt_command
    def acl_delete(self, name):
        if self.delete_acl(name):
            return "Deleted acl '%s'" % name
        return "ACL '%s' does not exist" % name

    @cmd(admin_only=True)
    @user_mgt_command
    def acl_allow(self, name, user):
        if self.add_user_to_allow(name, user):
            return "Added %s to %s (allow)" % (user, name)
        return "Failed to add %s to %s (allow)" % (user, name)

    @cmd(admin_only=True)
    @user_mgt_command
    def acl_deny(self, name, user):
        if self.add_user_to_deny(name, user):
            return "Added %s to %s (deny)" % (user, name)
        return "Failed to add %s to %s (deny)" % (user, name)

    @cmd(admin_only=True)
    @user_mgt_command
    def acl_remove(self, name, user):
        if self.remove_user_from_acl(name, user):
            return "Removed %s from %s (allow and deny)" % (user, name)
        return "Failed to remove %s from %s (allow and deny)" % (user, name)

    @cmd(admin_only=True)
    def acl_show(self, msg, args):
        """Show current allow and deny blocks for the given acl."""
        name = args[0] if len(args) > 0 else None
        if name is None:
            return "%s: The following ACLs are defined: %s" % (msg.user, ', '.join(self._acl.keys()))

        if name not in self._acl:
            return "Sorry, couldn't find an acl named '%s'" % name

        return '\n'.join([
            "%s: ACL '%s' is defined as follows:" % (msg.user, name),
            "allow: %s" % ', '.join(self._acl[name]['allow']),
            "deny: %s" % ', '.join(self._acl[name]['deny'])
        ])

    def add_user_to_allow(self, name, user):
        """Add a user to the given acl allow block."""

        # Clear user from both allow and deny before adding
        if not self.remove_user_from_acl(name, user):
            return False

        if name not in self._acl:
            return False

        self._acl[name]['allow'].append(user)
        return True

    def add_user_to_deny(self, name, user):
        """Add a user to the given acl deny block."""
        if not self.remove_user_from_acl(name, user):
            return False

        if name not in self._acl:
            return False

        self._acl[name]['deny'].append(user)
        return True

    def remove_user_from_acl(self, name, user):
        """Remove a user from the given acl (both allow and deny)."""
        if name not in self._acl:
            return False
        if user in self._acl[name]['allow']:
            self._acl[name]['allow'].remove(user)
        if user in self._acl[name]['deny']:
            self._acl[name]['deny'].remove(user)
        return True

    def create_acl(self, name):
        """Create a new acl."""
        if name in self._acl:
            return False

        self._acl[name] = {
            'allow': [],
            'deny': []
        }
        return True

    def delete_acl(self, name):
        """Delete an acl."""
        if name not in self._acl:
            return False

        del self._acl[name]
        return True

    @staticmethod
    def admin_check(cmd, user):
        # Commands not needing admin-level access pass
        if not cmd.admin_only:
            return True

        if hasattr(user, 'is_admin'):
            return user.is_admin
        return False

    def acl_check(self, cmd, user):
        effective_acl = cmd.acl
        if effective_acl not in self._acl:
            self.log.warn("Unable to locate ACL %s for %s, defaulting to *", effective_acl, cmd.method.__name__)
            effective_acl = '*'

        if self._check_allow(self._acl[effective_acl], user):
            return True
        if self._check_deny(self._acl[effective_acl], user):
            return False
        return True

    @staticmethod
    def _check_allow(acl, user):
        return '*' in acl['allow'] or user.username in acl['allow']

    @staticmethod
    def _check_deny(acl, user):
        return '*' in acl['deny'] or user.username in acl['deny']



from .room import SlackChannel, SlackGroup, SlackIM
from .user import SlackUser


class SlackEvent(object):
    """Encapsulates an event received from the RTM socket"""
    def __init__(self, sc=None, **kwargs):

        self._sc = sc
        self._channel = None
        self._user = None

        for k, v in kwargs.iteritems():
            setattr(self, k, v)

    @property
    def channel(self):
        return self._channel

    @channel.setter
    def channel(self, value):
        if isinstance(value, basestring):
            if value[0] == 'G':
                # Slack groups have an ID starting with 'G'
                self._channel = SlackGroup(value, sc=self._sc)
            elif value[0] == 'D':
                # Slack IMs have an ID starting with 'D'
                self._channel = SlackIM(value, sc=self._sc)
            else:
                self._channel = SlackChannel(value, sc=self._sc)
        else:
            self._channel = value

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, value):
        if isinstance(value, basestring):
            self._user = SlackUser(value, sc=self._sc)
        else:
            self._user = value



from .base import BasePlugin
from .manager import PluginManager


def cmd(admin_only=False, acl='*', aliases=None, while_ignored=False, *args, **kwargs):
    """
    Decorator to mark plugin functions as commands in the form of !<cmd_name>

    * admin_only - indicates only users in bot_admin are allowed to execute (only used if AuthManager is loaded)
    * acl - indicates which ACL to perform permission checks against (only used if AuthManager is loaded)
    * aliases - register function with additional commands (i.e. !alias1, !alias2, etc)
    * while_ignored - allows a command to be run, even if channel has been !sleep
    """
    def wrapper(func):
        func.is_cmd = True
        func.is_subcmd = len(func.__name__.split('_')) > 1
        func.cmd_name = func.__name__.replace('_', ' ')
        func.admin_only = admin_only
        func.acl = acl
        func.aliases = aliases
        func.while_ignored = while_ignored
        return func
    return wrapper


def webhook(*args, **kwargs):
    """
    Decorator to mark plugin functions as entry points for web calls

    * route - web route to register, uses Flask syntax
    * method - GET/POST, defaults to POST
    """
    def wrapper(func):
        func.is_webhook = True
        func.route = args[0]
        func.form_params = kwargs.get('form_params', [])
        func.method = kwargs.get('method', 'POST')
        return func
    return wrapper
